﻿using System;
using System.Web;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Threading;
using System.Net;
using System.Reflection;
using System.Collections.Concurrent;
using System.Linq.Expressions;
using System.Runtime.CompilerServices;
using System.Web.Script.Serialization;
using System.Runtime.Serialization;
using System.Dynamic;
using Microsoft.CSharp.RuntimeBinder;


namespace MapReduce
{
    ////////////////////////////////////////////////////
    // How to call from aspx page:
    //
    //    <script src="jquery-1.4.1.min.js" type="text/javascript"></script>
    //
    //    <script language="javascript" type="text/javascript">
    //        function Button1_onclick() {
    //            $(function () {
    //                $.ajax({
    //                    type: "GET",
    //                    url: 'http://localhost/MapReducePortal/BookSales?distributednodes=Node1,Node2,Node3&union=ISBN&sum=Sales&where=Sales>=10&sort=ISBN',
    //                    //url: 'http://localhost/MapReducePortal/BookSales?distributednodes=Node1,Node2,Node3&union=ISBN&sum=Sales&where=BookName==Book A,or BookName==Book C&sort=ISBN',
    //                    //url: 'http://localhost/MapReducePortal/Alarms?distributednodes=Node1,Node2,Node3&union=AlarmGuid&sum=AlarmDuration&where=Severity==Critical,and AlarmDuration>10&sort=CreateDate',
    //                    data: '',
    //                    contentType: 'application/json; charset=utf-8',
    //                    dataType: 'json',
    //                    success: function (data) {
    //                        var output = BuildTable(data);
    //                        $('#div1').html(output[0]);
    //
    //                        $('#divErrors').hide();
    //                        if (output[1] != null) {
    //                            $('#divErrors').show();
    //                            $('#divErrors').html(output[1]);
    //                        }
    //                    },
    //                    error: function (data) {
    //                        $('#b1').html(data.responseText);
    //                        alert("Error");
    //                    }
    //                });
    //            });
    //        }
    //    </script>
    //
    //    <html>
    //        <body>
    //            <h2>
    //                <input id="Button1" type="button" value="Test" onclick="return Button1_onclick()" />
    //            </h2>
    //            <br />
    //            <div id="div1"></div>
    //            <br />
    //            <div id="divErrors" style="display: none;"></div>
    //        </body>
    //    </html>
    //
    //
    // Web.config on IIS 7.0:
    //    <system.webServer>
    //      <modules>
    //        <add name="MapReduceModule" type="MapReduce.MapReduceModule" />
    //      </modules>
    //    </system.webServer>
    //
    // Put MapReduceModule.dll in the /bin directory of the virtual directory on IIS.
    //
    // To allow test simulation all on one node:
    // 1) Add virtual directory for: MapReduceData1, MapReduceData2, MapReduceData3
    // 2) Put MapReduceModule.dll in the /bin directory under: MapReduceData1, MapReduceData2, MapReduceData3
    // 3) Do ajax request using: Node1, Node2, Node3  (i.e. distributednodes=Node1,Node2,Node3)
    //
    // To debug:
    // 1) Open MapReduceModule project in Visual Studio
    // 2) Select 'Attach to Process...' in Debug menu in Visual Studio
    // 3) Attach to 'w3wp.exe'
    //    (Note: If no w3wp.exe, then run aspx page one time, retry, and w3wp.exe should now show in list
    //           If multiple w3wp.exe's show in list, then choose the one that is for MapReduceModule, not aspx page)
    //
    // Note: Make sure the IIS virtual directory is using an AppPool that is using .Net Framework 4.0.
    //       Define a new AppPool using .Net Framework 4.0, if no other AppPool exists that is using .Net Framework 4.0.
    //
    // Note: Can also set 'Maximum Worker Processes' to a number larger than the default of 1.
    //       This will allow there to be more w3wp.exe processes doing work in parallel.
    //       (Be careful if setting to more then 1 worker process because the static lists,
    //       dictionaries, etc will be destroyed and recreated when worker process recycles.)
    //       To change setting right click our AppPool and choose 'Advanced Settings'
    //          Go to 'Process Model' section and click 'Maximum Worker Processes' entry
    //             Change value from 1 to the new number of worker processes
    //
    // Note: Right now we only allow distributed "getting", and do not allow any distributed "setting".
    //       (i.e.  we only allow the Read of CRUD, we do not allow the Create, Update, Delete of CRUD)
    //       We only allow http method GET and POST. (POST does not indicate to create a new item;
    //       it only indicates the "chaining" of GET calls, and that the data in the request body
    //       is the results from a previous GET call that should be merged with the results of the
    //       current GET call)
    ////////////////////////////////////////////////////

    #region MapReduceModule

    public class MapReduceModule : IHttpModule
    {
        public void Dispose()
        {
        }
        public void Init(HttpApplication context)
        {
            context.AddOnBeginRequestAsync(new BeginEventHandler(OnBeginMap), new EndEventHandler(OnEndMap));
            context.AddOnEndRequestAsync(new BeginEventHandler(OnBeginReduce), new EndEventHandler(OnEndReduce));

            //load URI template table once at startup with all methods marked with PathInfoAttribute
            context.Application.Lock();
            if (context.Application["URITemplateTable"] == null)
            {
                //string baseUri = context.Context.Request.Url.Scheme + "://" + context.Context.Request.Url.Host + context.Context.Request.CurrentExecutionFilePath;
                string baseUri = context.Context.Request.Url.AbsoluteUri;
                context.Application["URITemplateTable"] = new UriTemplateTable(new Uri(baseUri));

                Type[] types = Assembly.GetExecutingAssembly().GetTypes();
                foreach (Type type in types)
                {
                    MethodInfo[] mis = type.GetMethods(BindingFlags.Static | BindingFlags.Public);
                    foreach (MethodInfo mi in mis)
                    {
                        object[] attributes = mi.GetCustomAttributes(typeof(BusinessObjects.PathInfoAttribute), false);
                        if (attributes.GetLength(0) > 0)
                        {
                            string attribReturnItemType = ((BusinessObjects.PathInfoAttribute)attributes[0]).ReturnItemType;
                            string atrribPathInfo = ((BusinessObjects.PathInfoAttribute)attributes[0]).PathInfo;
                            UriTemplate template = new UriTemplate(atrribPathInfo);
                            ((UriTemplateTable)context.Application["URITemplateTable"]).KeyValuePairs.Add(new KeyValuePair<UriTemplate, object>(template, new Tuple<MethodInfo, string>(mi, attribReturnItemType)));
                        }
                    }
                }
                ((UriTemplateTable)context.Application["URITemplateTable"]).MakeReadOnly(false);
            }
            context.Application.UnLock();
        }

        private IAsyncResult OnBeginMap(object sender, EventArgs e, AsyncCallback cb, object extraData)
        {
            HttpApplication application = (HttpApplication)sender;

            //force client calls to get new data every time
            application.Response.ExpiresAbsolute = DateTime.Now.Subtract(new TimeSpan(1, 0, 0, 0));
            application.Response.Expires = -1;

            //encoding
            application.Response.ContentType = "application/json";
            application.Response.ContentEncoding = Encoding.UTF8;

            //only allow http methods GET and POST
            if (application.Request.HttpMethod != "GET" && application.Request.HttpMethod != "POST")
            {
                application.Context.Response.Clear();
                application.Context.Response.ClearContent();
                application.Context.Response.StatusCode = 400;  //400 (BadRequest)
                application.Context.Response.SubStatusCode = int.MaxValue;  //custom indicator to skip the processing in the reduce step
                application.Context.Response.Write(JsonError(new Exception("Http method can only be GET or POST (PUT, DELETE, etc. are not implemented.")));
                application.Context.Response.Flush();
                application.Context.ApplicationInstance.CompleteRequest();
            }

            //thread-safe containers for url query string parameters, data, and errors
            ConcurrentDictionary<string, object> actions = new ConcurrentDictionary<string, object>();
            ConcurrentDictionary<string, object> data = new ConcurrentDictionary<string, object>();
            ConcurrentBag<string> errors = new ConcurrentBag<string>();

            //resulting data from previous call can be sent in as input data in post call
            //(used for chaining ajax calls to allow multiple levels (tiers) of mapreduce calls)
            if (application.Request.HttpMethod == "POST")
            {
                //byte[] bytes = application.Request.BinaryRead(application.Request.TotalBytes);
                //string input = application.Request.ContentEncoding.GetString(bytes);
                string input = new StreamReader(application.Request.InputStream).ReadToEnd();
                application.Request.InputStream.Position = 0;
                data.TryAdd("PostData", input);
            }

            //edge nodes that will be part of distributed request
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["DistributedNodes"]))
                actions.TryAdd("DistributedNodes", application.Context.Request.QueryString["DistributedNodes"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));

            //aggregations
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Sum"]))
                actions.TryAdd("Sum", application.Context.Request.QueryString["Sum"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));

            //operations
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Union"]))
                actions.TryAdd("Union", application.Context.Request.QueryString["Union"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));
            else if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Intersect"]))
                actions.TryAdd("Intersect", application.Context.Request.QueryString["Intersect"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));
            else if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Except"]))
                actions.TryAdd("Except", application.Context.Request.QueryString["Except"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));
            else if (!string.IsNullOrEmpty(application.Context.Request.QueryString["SymmetricExcept"]))
                actions.TryAdd("SymmetricExcept", application.Context.Request.QueryString["SymmetricExcept"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));

            //custom function
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Reduce"]))
                actions.TryAdd("Reduce", application.Context.Request.QueryString["Reduce"]);

            //filter
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Where"]))
                actions.TryAdd("Where", application.Context.Request.QueryString["Where"].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));

            //sort
            if (!string.IsNullOrEmpty(application.Context.Request.QueryString["Sort"]))
                actions.TryAdd("Sort", application.Context.Request.QueryString["Sort"]);
            else if (!string.IsNullOrEmpty(application.Context.Request.QueryString["SortAsc"]))
                actions.TryAdd("SortAsc", application.Context.Request.QueryString["SortAsc"]);
            else if (!string.IsNullOrEmpty(application.Context.Request.QueryString["SortDesc"]))
                actions.TryAdd("SortDesc", application.Context.Request.QueryString["SortDesc"]);

            //add containers to HttpContext
            lock (application.Context.Items.SyncRoot)
            {
                application.Context.Items.Add("#Actions#", actions);
                application.Context.Items.Add("#Data#", data);
                application.Context.Items.Add("#Errors#", errors);
            }

            //launch thread to do map requests without blocking pipeline
            WaitCallback wc = DoMap;
            return wc.BeginInvoke(application.Context, cb, application.Context);
        }
        private void OnEndMap(IAsyncResult ar)
        {
            System.Runtime.Remoting.Messaging.AsyncResult result = (System.Runtime.Remoting.Messaging.AsyncResult)ar;
            WaitCallback d = (WaitCallback)result.AsyncDelegate;
            d.EndInvoke(ar);
        }
        private void DoMap(object state)
        {
            HttpContext context = (HttpContext)state;

            //launch each map request on its own thread and wait until all are done to continue
            List<Tuple<string, HttpContext>> nodes = BuildNodesList(context);
            ParallelOptions options = new ParallelOptions();
            options.MaxDegreeOfParallelism = (int)10;  //make this a setting in the web.config
            Parallel.ForEach(nodes, options, new Action<Tuple<string, HttpContext>>(DoRequest));
        }
        private List<Tuple<string, HttpContext>> BuildNodesList(HttpContext context)
        {
            List<Tuple<string, HttpContext>> list = new List<Tuple<string, HttpContext>>();

            //get distributed edge nodes
            if (((ConcurrentDictionary<string, object>)context.Items["#Actions#"]).ContainsKey("DistributedNodes"))
            {
                string[] distributedNodes = (string[])((ConcurrentDictionary<string, object>)context.Items["#Actions#"])["DistributedNodes"];
                foreach (string nodeName in distributedNodes)
                {
                    list.Add(new Tuple<string, HttpContext>(nodeName, context));
                }
            }

            //always add self
            list.Add(new Tuple<string, HttpContext>("#Self#", context));

            return list;
        }
        private void DoRequest(Tuple<string, HttpContext> state)
        {
            //////////////////////////////////////////////////////////////////////////////////////////
            //This method is called here on the local node that initiated the request, and this exact
            //same method is called out on the edge nodes that were specified in the original request.
            //So when we are in this method we are either the local node or the edge node, and we need
            //to either invoke a local method (self) or invoke a web request (edge).
            //(e.g.  Self request = data method invoked on local node
            //       Not Self request = web request invoked to edge node)
            //
            //              <local-node>                       <edge-node>
            //               /       \                          /       \
            //    <self request>   <not self request>   <self request>   <not self request>
            //                                                           (not implemented)
            //
            //Right now we only allow ONE level of hierarchy because we remove any of the
            //'not self request' from the URL query string in the web request that we send to
            //the edge node. As a future enhancement we could leave them and allow MULTIPLE levels of
            //hierarchy. We could have a hierarchy of nodes within the same mesh, or a hierarchy of
            //meshes where meshes contain meshes. We could then make recursive calls on the hierarchy.
            //////////////////////////////////////////////////////////////////////////////////////////

            string uri = string.Empty;
            try
            {
                //build uri
                uri = state.Item2.Request.Url.Scheme + "://" + (state.Item1 == "#Self#" ? state.Item2.Request.Url.Host : state.Item1) + state.Item2.Request.Path;

                //check if edge node (defaults to false if not specified)
                bool edgeNode = (!string.IsNullOrEmpty(state.Item2.Request.QueryString["EdgeNode"]) && state.Item2.Request.QueryString["EdgeNode"].ToLower() == "true");

                //check if include local node data (defaults to true if not specified)
                bool includeLocalData = !(!string.IsNullOrEmpty(state.Item2.Request.QueryString["LocalData"]) && state.Item2.Request.QueryString["LocalData"].ToLower() == "false");

                
                if (state.Item1 == "#Self#")
                {
                    /////////////////////////////////////
                    //local method invoked on local node
                    /////////////////////////////////////
                    try
                    {
                        System.Web.Script.Serialization.JavaScriptSerializer js = new System.Web.Script.Serialization.JavaScriptSerializer();
                        js.RegisterConverters(new[] { new DynamicRowConverter(typeof(MapReduce.Row)) });
                        object result = null;

                        //check uri template table for method matching this request uri
                        UriTemplateMatch match = ((UriTemplateTable)state.Item2.Application["URITemplateTable"]).MatchSingle(new Uri(uri));
                        if (match != null)
                        {
                            Tuple<MethodInfo, string> data = (Tuple<MethodInfo, string>)match.Data;

                            //get parameters for method
                            ParameterInfo[] pis = data.Item1.GetParameters();
                            object[] parameters = new object[pis.GetLength(0)];
                            for (int i = 0; i < pis.GetLength(0); i++)
                            {
                                parameters[i] = js.ConvertToType(match.BoundVariables[pis[i].Name].ToString(), pis[i].ParameterType);
                            }

                            //store return item type that was specified in PathInfoAttribute
                            ((ConcurrentDictionary<string, object>)state.Item2.Items["#Actions#"]).TryAdd("ReturnItemType", data.Item2);

                            //execute method and get results
                            result = data.Item1.Invoke(null, parameters);
                        }
                        else
                            throw new Exception("Request URI does not match any URITemplate.");

                        //serialize results
                        StringBuilder sb = new StringBuilder();
                        js.Serialize(result, sb);
                        if (edgeNode)
                        {
                            //write results to response stream
                            state.Item2.Response.Write(sb.ToString());
                            state.Item2.Response.StatusCode = 200;
                            state.Item2.Response.SubStatusCode = int.MaxValue;  //custom indicator to skip the processing in the reduce step on this edge node
                            state.Item2.Response.Flush();
                            state.Item2.ApplicationInstance.CompleteRequest();  //jump directly to reduce step in the pipeline on this edge node
                        }
                        else
                        {
                            //store results in HttpContext for reduce step
                            if (includeLocalData)
                                ((ConcurrentDictionary<string, object>)state.Item2.Items["#Data#"]).TryAdd("LocalData", sb.ToString());
                        }
                    }
                    catch (Exception ex)
                    {
                        string error = JsonError(uri, ex);
                        //Note: Return status code 400 (BadRequest) from this edge node so it will trigger a
                        //WebException back on the local node that initiated the web request (if return 200 (OK)
                        //or 204 (NoContent), then will NOT trigger WebException)
                        state.Item2.Response.StatusCode = 400;
                        state.Item2.Response.SubStatusCode = int.MaxValue;  //custom indicator to skip the processing in the reduce step on this edge node
                        state.Item2.Response.Write(error);  //write error back as data in response stream along with 400 even though 400 will trigger WebException back at local node that initiated the web request
                        state.Item2.Response.Flush();  //need to flush here or will just return 404 (NotFound) because HttpModule only and no HttpHandler to handle request
                        state.Item2.ApplicationInstance.CompleteRequest();  //jump directly to reduce step in the pipeline on this edge node

                        //Note: Can choose whether to return 200 (OK) or 400 (BadRequest) here.
                        //      400 (BadRequest) will trigger WebException back on the local node that initiated the web request.
                        //      200 (OK) will NOT trigger WebException back on the local node that initiated the web request.
                        //      If return 200 (OK) then will need to parse the data that is returned from the web request looking for errors.
                        //      If return 400 (BadRequest) can catch WebException for errors.
                        //      After have errors back on local node that initiated web request, need to choose whether want to
                        //      return 200 (OK), 400 (BadRequest), or 500 (InternalServerError) from final reduce step back to the ajax call.
                        //      200 (OK) response will go to 'success:' handler of ajax call.
                        //      400 (BadRequest) or 500 (InternalServerError) response will go to 'error:' handler of ajax call.
                    }
                }
                else
                {
                    ///////////////////////////////////////////////
                    //web request invoked to distributed edge node
                    ///////////////////////////////////////////////
                    try
                    {
                        #region Testing - allows test simulation all on one node
                        //Need to add virtual directory for: MapReduceData1, MapReduceData2, MapReduceData3
                        //Call with ajax request using: Node1, Node2, Node3
                        uri = state.Item2.Request.Url.Scheme + "://localhost" + state.Item2.Request.Path;
                        if (state.Item1 == "Node1")
                            uri = uri.Replace("MapReducePortal", "MapReduceData1");
                        else if (state.Item1 == "Node2")
                            uri = uri.Replace("MapReducePortal", "MapReduceData2");
                        else if (state.Item1 == "Node3")
                            uri = uri.Replace("MapReducePortal", "MapReduceData3");
                        else
                            throw new Exception("Invalid node...");
                        #endregion


                        //mark as request to distributed edge node
                        uri = uri + "?EdgeNode=true";

                        //do web request to distributed edge node
                        HttpWebRequest webRequest = (HttpWebRequest)HttpWebRequest.Create(new Uri(uri));
                        webRequest.ContentType = "application/json; charset=utf-8";
                        webRequest.MediaType = "json";
                        webRequest.Timeout = Timeout.Infinite;
                        WebResponse response = webRequest.GetResponse();
                        using (StreamReader sm = new StreamReader(response.GetResponseStream()))
                        {
                            //get resulting data back from distributed edge node and put in HttpContext
                            string data = sm.ReadToEnd();
                            if (!string.IsNullOrEmpty(data))
                                ((ConcurrentDictionary<string, object>)state.Item2.Items["#Data#"]).TryAdd(uri, data);
                        }
                    }
                    catch (WebException ex)
                    {
                        string error = string.Empty;

                        //Note: This status code is 400 (BadRequest) because we set it to
                        //400 (BadRequest) in the error handler out on the edge node.
                        if (((HttpWebResponse)ex.Response).StatusCode == HttpStatusCode.BadRequest)
                        {
                            using (StreamReader sr = new StreamReader(ex.Response.GetResponseStream()))
                            {
                                error = sr.ReadToEnd();
                            }
                        }

                        if (error == string.Empty)
                            error = JsonError(uri, ex);

                        if (edgeNode)
                            throw new NotImplementedException("Multiple levels of hierarchy not implemented.");
                        else
                            ((ConcurrentBag<string>)state.Item2.Items["#Errors#"]).Add(error);
                    }
                    catch (Exception ex)
                    {
                        string error = JsonError(uri, ex);
                        if (edgeNode)
                            throw new NotImplementedException("Multiple levels of hierarchy not implemented.");
                        else
                            ((ConcurrentBag<string>)state.Item2.Items["#Errors#"]).Add(error);
                    }
                }
            }
            catch (Exception ex)
            {
                string error = JsonError(uri, ex);
                ((ConcurrentBag<string>)state.Item2.Items["#Errors#"]).Add(error);
            }
        }

        private IAsyncResult OnBeginReduce(object sender, EventArgs e, AsyncCallback cb, object extraData)
        {
            //launch thread to do reduce without blocking pipeline
            HttpApplication application = (HttpApplication)sender;
            WaitCallback wc = DoReduce;
            return wc.BeginInvoke(application.Context, cb, null);
        }
        private void OnEndReduce(IAsyncResult ar)
        {
            System.Runtime.Remoting.Messaging.AsyncResult result = (System.Runtime.Remoting.Messaging.AsyncResult)ar;
            WaitCallback d = (WaitCallback)result.AsyncDelegate;
            d.EndInvoke(ar);
        }
        private void DoReduce(object state)
        {
            HttpContext context = (HttpContext)state;

            //SubStatusCode=int.MaxValue is our custom indicator to skip procesing in reduce step
            if (context.Response.SubStatusCode == int.MaxValue)
                return;

            try
            {
                //Note: Status code at this point is usually 404 (NotFound) because there is no HttpHandler,
                //we change it and return 200 (OK) for everything and just include the errors in the data.
                context.Response.StatusCode = 200;

                //do reduce
                if (context.Items.Count > 0)
                {
                    Hashtable ht = (Hashtable)context.Items;
                    ConcurrentDictionary<string, object> actions = (ConcurrentDictionary<string, object>)ht["#Actions#"];
                    ConcurrentDictionary<string, object> data = (ConcurrentDictionary<string, object>)ht["#Data#"];
                    ConcurrentBag<string> errors = (ConcurrentBag<string>)ht["#Errors#"];

                    //aggregations
                    string[] sumKeys = null;
                    if (actions.ContainsKey("Sum"))
                        sumKeys = (string[])actions["Sum"];

                    //operations
                    string[] keys = null;
                    if (actions.ContainsKey("Union"))
                        keys = (string[])actions["Union"];
                    else if (actions.ContainsKey("Intersect"))
                        keys = (string[])actions["Intersect"];
                    else if (actions.ContainsKey("Except"))
                        keys = (string[])actions["Except"];
                    else if (actions.ContainsKey("SymmetricExcept"))
                        keys = (string[])actions["SymmetricExcept"];

                    //get return type of individual item in results
                    string returnItemType = string.Empty;
                    if (actions.ContainsKey("ReturnItemType"))
                        returnItemType = (string)actions["ReturnItemType"];
                    if (string.IsNullOrEmpty(returnItemType))
                        throw new Exception("ReturnItemType not specified.");
                    Type typeReturnItem = Type.GetType(returnItemType);

                    //custom function
                    MethodInfo miCustomfunction = null;
                    if (actions.ContainsKey("Reduce"))
                    {
                        string customFunctionName = (string)actions["Reduce"];
                        Type[] types = Assembly.GetExecutingAssembly().GetTypes();
                        foreach (Type type in types)
                        {
                            MethodInfo mi = type.GetMethod(customFunctionName, BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.Public);
                            if (mi != null)
                            {
                                //make sure method is marked with custom reduce function attribute
                                object[] attributes = mi.GetCustomAttributes(typeof(BusinessObjects.CustomReduceFunctionAttribute), false);
                                if (attributes.GetLength(0) == 0)
                                    throw new Exception("Custom reduce function specified [" + customFunctionName + "] has not been marked with [CustomReduceFunction] attribute.");

                                //make sure signature is correct, it must be: static HashSet<MapReduce.Row> CustomFunctionName(HashSet<MapReduce.Row> currentData, HashSet<MapReduce.Row> newData)
                                ParameterInfo[] pis = mi.GetParameters();
                                if (!mi.IsStatic || pis.GetLength(0) != 2 || mi.ReturnType != typeof(HashSet<MapReduce.Row>) || pis[0].ParameterType != typeof(HashSet<MapReduce.Row>) || pis[1].ParameterType != typeof(HashSet<MapReduce.Row>))
                                    throw new Exception("Custom function signature must be:  static HashSet[Row] CustomFunctionName(HashSet[Row] currentData, HashSet[Row] newData)");

                                miCustomfunction = mi;
                                break;
                            }
                        }
                        if (miCustomfunction == null)
                            throw new Exception("Custom function [" + customFunctionName + "] not found in this assembly.");
                    }


                    System.Web.Script.Serialization.JavaScriptSerializer js = new System.Web.Script.Serialization.JavaScriptSerializer();
                    js.RegisterConverters(new[] { new DynamicRowConverter(typeReturnItem) });

                    //get data from previous ajax request if was sent as input to this request
                    HashSet<Row> postResults = null;
                    if (data.ContainsKey("PostData"))
                    {
                        dynamic tempPostResults = js.Deserialize((string)data["PostData"], typeof(HashSet<Row>));
                        postResults = new HashSet<Row>(tempPostResults);
                    }

                    //create final rows object that will be returned  (put in post results if there are any)
                    HashSet<Row> rows = null;
                    if (postResults != null)
                        rows = new HashSet<Row>(postResults, new ObjectEqualityComparer(keys, sumKeys));
                    else
                        rows = new HashSet<Row>(new ObjectEqualityComparer(keys, sumKeys));

                    //get local data and put in final rows object first
                    if (data.ContainsKey("LocalData"))
                    {
                        dynamic tempLocalResults = js.Deserialize((string)data["LocalData"], typeof(HashSet<Row>));
                        HashSet<Row> localResults = new HashSet<Row>(tempLocalResults);
                        foreach (Row item in localResults)
                        {
                            rows.Add(item);
                        }

                        //if custom function specified then execute custom function against local data
                        if (rows.Count > 0)
                        {
                            if (miCustomfunction != null)
                            {
                                HashSet<Row> currentRows = new HashSet<Row>();
                                if (keys != null && keys.GetLength(0) > 0 && sumKeys != null && sumKeys.GetLength(0) > 0)
                                    currentRows = new HashSet<Row>(new ObjectEqualityComparer(keys, sumKeys));
                                else if (keys != null && keys.GetLength(0) > 0)
                                    currentRows = new HashSet<Row>(new ObjectEqualityComparer(keys));

                                HashSet<Row> newRows = new HashSet<Row>(rows);
                                rows = (HashSet<Row>)miCustomfunction.Invoke(rows, new object[] { currentRows, newRows });
                            }
                        }
                    }

                    //get data results from all the ajax calls to all the edge nodes
                    //and merge them into final rows object
                    foreach (object key in data.Keys)
                    {
                        if (key != null && (string)key != "LocalData" && (string)key != "PostData")
                        {
                            string url = key.ToString();
                            string nodeData = (string)data[url];

                            //either perform custom reduce function or perform operation like union, intersect, etc.
                            dynamic nodeResults = js.Deserialize(nodeData, typeof(HashSet<Row>));
                            HashSet<Row> newRows = new HashSet<Row>(nodeResults);
                            if (miCustomfunction != null)
                            {
                                //custom reduce function
                                rows = (HashSet<Row>)miCustomfunction.Invoke(null, new object[] { rows, newRows });
                            }
                            else
                            {
                                //operations
                                if (actions.ContainsKey("Union"))
                                    rows.UnionWith(newRows);
                                else if (actions.ContainsKey("Intersect"))
                                    rows.IntersectWith(newRows);
                                else if (actions.ContainsKey("Except"))
                                    rows.ExceptWith(newRows);
                                else if (actions.ContainsKey("SymmetricExcept"))
                                    rows.SymmetricExceptWith(newRows);
                                else
                                    rows.UnionWith(newRows);
                            }
                        }
                    }

                    //prepare final results to be returned in response
                    if (rows.Count > 0)
                    {
                        //filter (where clause)
                        if (actions.ContainsKey("Where"))
                        {
                            string[] whereExpressions = (string[])actions["Where"];
                            if (whereExpressions != null && whereExpressions.GetLength(0) > 0)
                            {
                                string[] propertyNames = null;
                                dynamic p = ReversePredicateBuilder(rows, whereExpressions, typeof(Row), out propertyNames);
                                HashSet<Row> whereRows = new HashSet<Row>(rows, new WhereObjectEqualityComparer(propertyNames));
                                whereRows.RemoveWhere(p);
                                //rows IEqualityComparer is lost when assign whereRows, but ok because we don't
                                //do anymore operations except serialize on rows object after this point before
                                //returning results to client
                                rows = whereRows;
                            }
                        }

                        string sortField = string.Empty;
                        bool sortAscending = true;
                        if (actions.ContainsKey("Sort"))
                            sortField = (string)actions["Sort"];
                        else if (actions.ContainsKey("SortAsc"))
                            sortField = (string)actions["SortAsc"];
                        else if (actions.ContainsKey("SortDesc"))
                        {
                            sortField = (string)actions["SortDesc"];
                            sortAscending = false;
                        }

                        //serialize final results
                        StringBuilder sb = new StringBuilder();
                        if (!string.IsNullOrEmpty(sortField))
                        {
                            //sort
                            List<Row> sortedRows = new List<Row>(rows);
                            sortedRows.Sort(new ObjectComparer(sortField, sortAscending));
                            js.Serialize(sortedRows, sb);
                        }
                        else
                            js.Serialize(rows, sb);

                        //if any errors, build errors envelope object around all errors
                        //and append to end of data just like another row of data
                        AppendJsonErrors(sb, errors);

                        //write response to response stream
                        context.Response.Write(sb.ToString());
                    }
                    else
                    {
                        //if any errors, write errors to response stream (even if no data)
                        if (errors.Count > 0)
                        {
                            string temp = JsonError(errors);
                            context.Response.Write(temp);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                context.Response.Clear();
                context.Response.ClearContent();
                context.Response.StatusCode = 400;  //returning 400 (BadRequest) response will go to 'error:' handler of ajax call.
                string error = JsonError(ex);
                //error = context.Server.HtmlEncode(error);
                context.Response.Write(error);
            }
            finally
            {
                context.Response.Flush();
                context.Response.End();
            }
        }

        
        #region Helpers

        private string JsonError(params Exception[] exceptions)
        {
            return JsonError(null, exceptions);
        }
        private string JsonError(string uri, params Exception[] exceptions)
        {
            if (exceptions.GetLength(0) == 0)
                throw new ArgumentException("Must supply at least one Exception.", "exceptions");

            if (exceptions.GetLength(0) == 1)
            {
                if (string.IsNullOrEmpty(uri))
                    return "{\"Error\":\"" + exceptions[0].GetBaseException().Message + "\"}";
                else
                    return "{\"Error\":\"" + exceptions[0].GetBaseException().Message + "\", \"URI\":\"" + uri + "\"}";
            }

            StringBuilder sb = new StringBuilder();
            sb.Append("{\"Errors\":[");
            foreach (Exception ex in exceptions)
            {
                string error = string.Empty;
                if (string.IsNullOrEmpty(uri))
                    error = "{\"Error\":\"" + ex.GetBaseException().Message + "\"}";
                else
                    error = "{\"Error\":\"" + ex.GetBaseException().Message + "\", \"URI\":\"" + uri + "\"}";
                sb.Append(error);
                sb.Append(',');
            }
            if (sb[sb.Length-1] == ',')
                sb.Replace(",", "]}", sb.Length - 1, 1);
            else
                sb.Append("]}");

            return sb.ToString();
        }
        private string JsonError(ConcurrentBag<string> errors)
        {
            if (errors.Count == 0)
                throw new ArgumentException("Must supply at least one error.", "errors");

            StringBuilder sb = new StringBuilder();
            sb.Append("{\"Errors\":[");
            foreach (string error in errors)
            {
                sb.Append(error);
                sb.Append(',');
            }
            if (sb[sb.Length - 1] == ',')
                sb.Replace(",", "]}", sb.Length - 1, 1);
            else
                sb.Append("]}");

            return sb.ToString();
        }

        private void AppendJsonErrors(StringBuilder sb, ConcurrentBag<string> errors)
        {
            if (errors.Count > 0)
            {
                bool data = sb.Length > 0;
                if (data)
                {
                    sb.Remove(sb.Length - 1, 1);
                    sb.Append(',');
                }
                sb.Append("{\"Errors\":[");
                foreach (string error in errors)
                {
                    sb.Append(error);
                    sb.Append(',');
                }
                
                if (sb[sb.Length - 1] == ',')
                    sb.Replace(",", "]}", sb.Length - 1, 1);
                else
                    sb.Append("]}");

                if (data)
                    sb.Append(']');
            }
        }

        private static dynamic ReversePredicateBuilder(HashSet<Row> instance, string[] expressions, Type predicateObjectType, out string[] propertyNames)
        {
            //Builds a predicate from a string.
            //Allows multiple expressions that are delimited by a comma, and have an 'and' or an 'or' between them.
            //(i.e. "BookName == Book A, and ID == 1, and Sales == 10")
            //
            //Example:
            //Pass in this:  "BookName == Book A, and ID == 1, and Sales == 10"
            //Builds predicate like this:  b => (((b.BookName != "Book A") Or (b.ID != Parse("1"))) Or (b.Sales != Parse("10")))
            //
            //Used to build a Predicate from a string passed as a query string parameter in a URL:
            //http://.../BookSales?where=BookName==Book A,and ID==1,and Sales==10
            //
            //Note: Reverses the operator and operation because this Predicate gets passed to
            //HashSet.RemoveWhere instead of just a Where. (i.e. RemoveWhere vs. Where)

            if (expressions == null || expressions.GetLength(0) == 0)
                throw new ArgumentException("No expressions supplied for 'where' parameter in URL query string.", "expressions");
            if (instance == null || instance.Count == 0)
                throw new ArgumentException("Result set was empty or null. Result set must have records for 'where' parameter in URL query string to build valid Predicate.", "instance");

            //build expression
            var p = Expression.Parameter(predicateObjectType, "b");
            var predicateType = typeof(Predicate<>).MakeGenericType(predicateObjectType);
            BinaryExpression body = null;
            List<string> propertyNamesList = new List<string>();

            foreach (string expression in expressions)
            {
                if (expression.Contains("=="))
                {
                    string[] parts = expression.Split(new string[] { "==" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), "==", parts[1].Trim(), ref propertyNamesList);
                }
                else if (expression.Contains("!="))
                {
                    string[] parts = expression.Split(new string[] { "!=" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), "!=", parts[1].Trim(), ref propertyNamesList);
                }
                else if (expression.Contains("<="))
                {
                    string[] parts = expression.Split(new string[] { "<=" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), "<=", parts[1].Trim(), ref propertyNamesList);
                }
                else if (expression.Contains(">="))
                {
                    string[] parts = expression.Split(new string[] { ">=" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), ">=", parts[1].Trim(), ref propertyNamesList);
                }
                else if (expression.Contains("<"))
                {
                    string[] parts = expression.Split(new string[] { "<" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), "<", parts[1].Trim(), ref propertyNamesList);
                }
                else if (expression.Contains(">"))
                {
                    string[] parts = expression.Split(new string[] { ">" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.GetLength(0) != 2) throw new Exception("Expression [" + expression + "] is invalid.");
                    body = ExpressionBuilder(instance, p, body, parts[0].Trim(), ">", parts[1].Trim(), ref propertyNamesList);
                }
                else
                    throw new Exception("No valid operator in expression for 'where' parameter in URL query string.");
            }

            propertyNames = propertyNamesList.ToArray();

            //create predicate
            var predicate = Expression.Lambda(predicateType, body, p);
            dynamic del = predicate.Compile();
            return del;
        }
        private static BinaryExpression ExpressionBuilder(HashSet<Row> instance, ParameterExpression p, BinaryExpression existingExpression, string propertyName, string operatorString, string propertyValue, ref List<string> propertyNames)
        {
            if (string.IsNullOrEmpty(operatorString))
                throw new ArgumentException("Must supply valid operator string.", operatorString);

            BinaryExpression newExpression = null;

            //determine append operation
            string operation = string.Empty;
            if (propertyName.StartsWith("and "))
            {
                operation = "and";
                propertyName = propertyName.Replace("and ", "");
            }
            else if (propertyName.StartsWith("or "))
            {
                operation = "or";
                propertyName = propertyName.Replace("or ", "");
            }

            //store all property names to return
            propertyNames.Add(propertyName);

            //get property type
            Row[] rows = new Row[1];
            instance.CopyTo(rows, 0, 1);
            Type propertyType = rows[0].GetPropertyType(propertyName);

            //build expression
            MethodInfo miGetPropertyValue = typeof(Row).GetMethod("GetPropertyValue", BindingFlags.Public | BindingFlags.Instance);
            var propertyGetValue = Expression.Call(p, miGetPropertyValue, Expression.Constant(propertyName));
            var property = Expression.Convert(propertyGetValue, propertyType);
            var value = Expression.Constant(propertyValue, typeof(string));
            MethodInfo mi = propertyType.GetMethod("Parse", BindingFlags.Static | BindingFlags.Public, null, new Type[] { typeof(string) }, null);
            if (mi != null)
            {
                var parse = Expression.Call(mi, value);
                if (operatorString == "==")
                    newExpression = Expression.NotEqual(property, parse);
                else if (operatorString == "!=")
                    newExpression = Expression.Equal(property, parse);
                else if (operatorString == "<=")
                    newExpression = Expression.GreaterThan(property, parse);
                else if (operatorString == ">=")
                    newExpression = Expression.LessThan(property, parse);
                else if (operatorString == "<")
                    newExpression = Expression.GreaterThanOrEqual(property, parse);
                else if (operatorString == ">")
                    newExpression = Expression.LessThanOrEqual(property, parse);
            }
            else
            {
                if (operatorString == "==")
                    newExpression = Expression.NotEqual(property, value);
                else if (operatorString == "!=")
                    newExpression = Expression.Equal(property, value);
                else if (operatorString == "<=")
                    newExpression = Expression.GreaterThan(property, value);
                else if (operatorString == ">=")
                    newExpression = Expression.LessThan(property, value);
                else if (operatorString == "<")
                    newExpression = Expression.GreaterThanOrEqual(property, value);
                else if (operatorString == ">")
                    newExpression = Expression.LessThanOrEqual(property, value);
            }

            //append expression
            if (existingExpression == null)
                return newExpression;
            else
            {
                if (operation == "and")
                    return Expression.Or(existingExpression, newExpression);
                else if (operation == "or")
                    return Expression.And(existingExpression, newExpression);
                else
                    return Expression.Or(existingExpression, newExpression);
            }
        }

        #endregion
    }

    #endregion

    
    
    #region Dynamic Row Object

    [Serializable()]
    public class Row : DynamicObject, ISerializable
    {
        private Dictionary<string, object> _properties = new Dictionary<string, object>();

        public Row()
        {
        }
        public Row(IDictionary<string, object> properties)
        {
            _properties = (Dictionary<string, object>)properties;
        }
        public Row(IDictionary<string, object> properties, Type type, JavaScriptSerializer serializer)
        {
            Dictionary<string, object> convertedDictionary = new Dictionary<string, object>();
            foreach (string key in properties.Keys)
            {
                convertedDictionary.Add(key, serializer.ConvertToType(properties[key], type.GetProperty(key).PropertyType));
            }
            _properties = convertedDictionary;
        }
        public override bool TryGetIndex(GetIndexBinder binder, object[] indexes, out object result)
        {
            if (_properties.ContainsKey(indexes[0].ToString()))
            {
                result = _properties[indexes[0].ToString()];
                return true;
            }
            else
            {
                result = null;
                return false;
            }
        }
        public override bool TrySetIndex(SetIndexBinder binder, object[] indexes, object value)
        {
            if (!_properties.ContainsKey(indexes[0].ToString()))
            {
                _properties.Add(indexes[0].ToString(), value);
                return true;
            }
            else
            {
                _properties[indexes[0].ToString()] = value;
                return true;
            }
        }
        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            if (_properties.ContainsKey(binder.Name))
            {
                result = _properties[binder.Name];
                return true;
            }
            else
            {
                result = null;
                return false;
            }
        }
        public override bool TrySetMember(SetMemberBinder binder, object value)
        {
            if (!_properties.ContainsKey(binder.Name))
            {
                _properties.Add(binder.Name, value);
                return true;
            }
            else
            {
                _properties[binder.Name] = value;
                return true;
            }
        }
        public override bool TryConvert(ConvertBinder binder, out object result)
        {
            //Explicit cast operations trigger this method.
            //(i.e. an explicit cast from dynamic object to a concrete type,
            //like explicit cast from Row to Book)
            dynamic newType = Activator.CreateInstance(binder.Type);
            if (newType == null)
            {
                result = newType;
                return false;
            }
            else
            {
                foreach (PropertyInfo pi in binder.Type.GetProperties(BindingFlags.Public | BindingFlags.Instance))
                {
                    pi.SetValue(newType, ((dynamic)this)[pi.Name], null);
                }
                result = newType;
                return true;
            }
        }

        public Dictionary<string, object> GetProperties()
        {
            return _properties;
        }
        public IEnumerable<string> GetPropertyNames()
        {
            foreach (string propertyName in _properties.Keys)
            {
                yield return propertyName;
            }
        }
        public object GetPropertyValue(string propertyName)
        {
            if (!_properties.ContainsKey(propertyName))
                throw new Exception("Property [" + propertyName + "] does not exist on object [Row].");

            return _properties[propertyName];
        }
        public Type GetPropertyType(string propertyName)
        {
            if (!_properties.ContainsKey(propertyName))
                throw new Exception("Property [" + propertyName + "] does not exist on object [Row].");

            return _properties[propertyName].GetType();
        }
        public bool IsPropertyExist(string propertyName)
        {
            return _properties.ContainsKey(propertyName);
        }

        public static void ConvertList(List<Row> sourceList, dynamic targetList)
        {
            targetList.Clear();

            //Converts list of dynamic objects into list of concrete type objects.
            //The dynamic object will be converted into whatever type the target list holds.
            Type type = targetList.GetType().GetGenericArguments()[0];
            foreach (dynamic item in sourceList)
            {
                var newItem = DynamicHelper.ExplicitCast(item, type);
                targetList.Add(newItem);
            }
        }

        #region Serialization
        public Row(SerializationInfo info, StreamingContext context)
        {
            foreach (var key in info)
            {
                dynamic item = this;
                item[key.Name] = key.Value;
            }
        }
        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            foreach (string property in _properties.Keys)
            {
                info.AddValue(property, _properties[property]);
            }
        }
        #endregion
    }
    public class DynamicRowConverter : JavaScriptConverter
    {
        private Type _schemaType = null;
        private Dictionary<string, Type> _types = new Dictionary<string, Type>();

        public DynamicRowConverter(Type schemaType)
        {
            _schemaType = schemaType;

            foreach (PropertyInfo pi in _schemaType.GetProperties(BindingFlags.Public | BindingFlags.Instance))
            {
                _types.Add(pi.Name, pi.PropertyType);
            }
        }
        private DynamicRowConverter()
        {
        }
        public override object Deserialize(IDictionary<string, object> dictionary, Type type, JavaScriptSerializer serializer)
        {
            if (type == typeof(Row))
            {
                Dictionary<string, object> newDictionary = new Dictionary<string, object>();
                foreach (string key in dictionary.Keys)
                {
                    if (_types.ContainsKey(key))
                        newDictionary.Add(key, serializer.ConvertToType(dictionary[key], _types[key]));
                    else
                        newDictionary.Add(key, dictionary[key]);
                }
                Row row = new Row(newDictionary);
                return row;
            }
            else
                throw new Exception("Type not supported.");
        }

        public override IDictionary<string, object> Serialize(object obj, JavaScriptSerializer serializer)
        {
            if (obj is Row)
            {
                IDictionary<string, object> temp = ((Row)obj).GetProperties();
                return temp;
            }
            else
                throw new Exception("Type not supported.");
        }

        public override IEnumerable<Type> SupportedTypes
        {
            get { return new[] { typeof(Row) }; }
        }
    }
    public static class DynamicHelper
    {
        private static ConcurrentDictionary<Type, object> _castCache = null;

        static DynamicHelper()
        {
            _castCache = new ConcurrentDictionary<Type, object>();
        }

        public static dynamic ExplicitCast(dynamic item, Type castTo)
        {
            //This utility method casts a custom dynamic type (that implements TryConvert) to a concrete type using type variables at runtime (instead of generic types at compile time).
            //It basically triggers an explicit cast that in turn triggers a call to TryConvert on the custom dynamic object.
            //This utility method caches the callsite used for casting the custom dynamic type based on the castTo type as a key.
            //Note: If custom dynamic type has more properties than concrete type, cast will still work, but it will only fill in propeties on concrete type and disregard the additonal properties on the custom dynamic type.
            //Note: If custom dynamic type does not have a property that exists on concrete type, cast will fail.
            //Note: Only use this to cast from a custom dynamic type (that implements TryConvert) to a regular concrete type.
            if (_castCache.ContainsKey(castTo))
            {
                dynamic callsite = _castCache[castTo];
                var val = callsite.Target(callsite, item);
                return val;
            }
            else
            {
                Type t = typeof(Func<,,>);
                Type typeCallSite = typeof(CallSite);
                Type typeItem = item.GetType();
                Type typeFunc = t.MakeGenericType(typeCallSite, typeItem, castTo);
                var binder = Microsoft.CSharp.RuntimeBinder.Binder.Convert(CSharpBinderFlags.ConvertExplicit, castTo, null);
                dynamic callsite = CallSite.Create(typeFunc, binder);
                var val = callsite.Target(callsite, item);

                _castCache.TryAdd(castTo, callsite);

                return val;
            }
        }
    }

    #endregion

    
    
    #region Comparers

    internal class ObjectEqualityComparer : IEqualityComparer<Row>
    {
        private string[] _keyStrings = null;
        private string[] _sumKeyStrings = null;

        public ObjectEqualityComparer(string[] keys)
        {
            _keyStrings = keys;
        }
        public ObjectEqualityComparer(string[] keys, string[] sumKeys)
        {
            _keyStrings = keys;
            _sumKeyStrings = sumKeys;
        }

        public bool Equals(Row x, Row y)
        {
            if (_keyStrings == null || _keyStrings.GetLength(0) == 0)
                return false;

            dynamic dx = x;
            dynamic dy = y;
            foreach (string key in _keyStrings)
            {
                if (!x.IsPropertyExist(key))
                    throw new Exception("Property [" + key + "] does not exist on object [Row].");

                int result = (int)dx[key].CompareTo(dy[key]);
                if (result != 0)
                    return false;
            }
            if (_sumKeyStrings != null)
            {
                foreach (string key in _sumKeyStrings)
                {
                    if (!x.IsPropertyExist(key))
                        throw new Exception("Property [" + key + "] does not exist on object [Row].");

                    dynamic sumValue = dx[key] + dy[key];
                    dx[key] = sumValue;
                }
            }
            return true;
        }

        public int GetHashCode(Row obj)
        {
            if (_keyStrings == null || _keyStrings.GetLength(0) == 0)
                return 0;

            dynamic d = obj;
            StringBuilder sb = new StringBuilder();
            foreach (string key in _keyStrings)
            {
                if (!obj.IsPropertyExist(key))
                    throw new Exception("Property [" + key + "] does not exist on object [Row].");

                sb.Append(d[key].ToString());
            }
            return sb.ToString().GetHashCode();
        }
    }

    internal class ObjectComparer : IComparer<Row>
    {
        private string _sortField = string.Empty;
        private bool _ascending = true;

        public ObjectComparer(string sortField)
        {
            _sortField = sortField;
        }
        public ObjectComparer(string sortField, bool ascending)
        {
            _sortField = sortField;
            _ascending = ascending;
        }
        private ObjectComparer()
        {
        }

        public int Compare(Row x, Row y)
        {
            int result = 0;

            if (!string.IsNullOrEmpty(_sortField))
            {
                dynamic dx = x;
                dynamic dy = y;

                if (!x.IsPropertyExist(_sortField))
                    throw new Exception("Property [" + _sortField + "] does not exist on object [Row].");

                if (dx[_sortField] == null && dy[_sortField] == null)
                    return 0;

                if (_ascending)
                {
                    if (dx[_sortField] == null)
                        return -1;
                    if (dy[_sortField] == null)
                        return 1;

                    return (int)dx[_sortField].CompareTo(dy[_sortField]);
                }
                else
                {
                    if (dx[_sortField] == null)
                        return 1;
                    if (dy[_sortField] == null)
                        return -1;

                    return (int)dy[_sortField].CompareTo(dx[_sortField]);
                }
            }
            return result;
        }
    }

    internal class WhereObjectEqualityComparer : IEqualityComparer<Row>
    {
        private string[] _whereStrings = null;

        public WhereObjectEqualityComparer(string[] whereKeys)
        {
            _whereStrings = whereKeys;
        }

        public bool Equals(Row x, Row y)
        {
            if (_whereStrings == null || _whereStrings.GetLength(0) == 0)
                return false;

            dynamic dx = x;
            dynamic dy = y;
            foreach (string key in _whereStrings)
            {
                if (!x.IsPropertyExist(key))
                    throw new Exception("Property [" + key + "] does not exist on object [Row].");

                int result = (int)dx[key].CompareTo(dy[key]);
                if (result != 0)
                    return false;
            }
            return true;
        }

        public int GetHashCode(Row obj)
        {
            if (_whereStrings == null || _whereStrings.GetLength(0) == 0)
                return 0;

            dynamic d = obj;
            StringBuilder sb = new StringBuilder();
            foreach (string key in _whereStrings)
            {
                if (!obj.IsPropertyExist(key))
                    throw new Exception("Property [" + key + "] does not exist on object [Row].");

                sb.Append(d[key].ToString());
            }
            return sb.ToString().GetHashCode();
        }
    }

    #endregion
}



#region BusinessObjects - Put all the business data objects here...

namespace BusinessObjects
{
    #region Attributes to use on business data objects

    [AttributeUsage(AttributeTargets.Method, AllowMultiple = false)]
    public class PathInfoAttribute : Attribute
    {
        public string PathInfo { get; set; }
        public string ReturnItemType { get; set; }
    }
    [AttributeUsage(AttributeTargets.Method, AllowMultiple = false)]
    public class CustomReduceFunctionAttribute : Attribute
    {
    }

    #endregion

    
    #region BookSales object - Used for testing

    [Serializable()]
    public class Book
    {
        public int ID { get; set; }
        public string BookName { get; set; }
        public Guid ISBN { get; set; }
        public int Sales { get; set; }
        public DateTime PublishDate { get; set; }
    }

    [Serializable()]
    public class BookSales : ReadOnlyCollectionBase
    {
        ////////////////////////////////////////
        //Example to test operations (like union, sum, etc) and sorting
        ////////////////////////////////////////
        [PathInfo(PathInfo = "/BookSales", ReturnItemType = "BusinessObjects.Book")]
        public static BookSales Get()
        {
            BookSales items = new BookSales();
            if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReducePortal"))
            {
                ////throw new Exception("test error on LocalNode...");
                items.InnerList.Add(new Book { ID = 1, BookName = "Book A", ISBN = new Guid("11111111-1111-1111-1111-111111111111"), Sales = 10, PublishDate = new DateTime(2011, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Book { ID = 2, BookName = "Book B", ISBN = new Guid("22222222-2222-2222-2222-222222222222"), Sales = 30, PublishDate = new DateTime(2012, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Book { ID = 3, BookName = "Book C", ISBN = new Guid("33333333-3333-3333-3333-333333333333"), Sales = 20, PublishDate = new DateTime(2013, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData1"))
            {
                ////throw new Exception("test error on Node1...");
                items.InnerList.Add(new Book { ID = 4, BookName = "Book D", ISBN = new Guid("44444444-4444-4444-4444-444444444444"), Sales = 5, PublishDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData2"))
            {
                ////throw new Exception("test error on Node2...");
                items.InnerList.Add(new Book { ID = 5, BookName = "Book E", ISBN = new Guid("55555555-5555-5555-5555-555555555555"), Sales = 20, PublishDate = new DateTime(2015, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData3"))
            {
                ////throw new Exception("test error on Node3...");
                items.InnerList.Add(new Book { ID = 3, BookName = "Book C", ISBN = new Guid("33333333-3333-3333-3333-333333333333"), Sales = 5, PublishDate = new DateTime(2013, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Book { ID = 4, BookName = "Book D", ISBN = new Guid("44444444-4444-4444-4444-444444444444"), Sales = 1, PublishDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            return items;
        }

        [PathInfo(PathInfo = "/BookSales/GetTotalBookSalesByBookName/{bookName}", ReturnItemType = "BusinessObjects.Book")]
        public static BookSales GetTotalBookSalesByBookName(string bookName)
        {
            BookSales items = new BookSales();
            if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReducePortal"))
            {
                if (bookName == "Book A")
                    items.InnerList.Add(new Book { ID = 1, BookName = "Book A", ISBN = new Guid("11111111-1111-1111-1111-111111111111"), Sales = 10, PublishDate = new DateTime(2011, 4, 1, 0, 0, 0, 0) });
                else if (bookName == "Book B")
                    items.InnerList.Add(new Book { ID = 2, BookName = "Book B", ISBN = new Guid("22222222-2222-2222-2222-222222222222"), Sales = 30, PublishDate = new DateTime(2012, 4, 1, 0, 0, 0, 0) });
                else if (bookName == "Book C")
                    items.InnerList.Add(new Book { ID = 3, BookName = "Book C", ISBN = new Guid("33333333-3333-3333-3333-333333333333"), Sales = 20, PublishDate = new DateTime(2013, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData1"))
            {
                if (bookName == "Book D")
                    items.InnerList.Add(new Book { ID = 4, BookName = "Book D", ISBN = new Guid("44444444-4444-4444-4444-444444444444"), Sales = 5, PublishDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData2"))
            {
                if (bookName == "Book E")
                    items.InnerList.Add(new Book { ID = 5, BookName = "Book E", ISBN = new Guid("55555555-5555-5555-5555-555555555555"), Sales = 20, PublishDate = new DateTime(2015, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData3"))
            {
                if (bookName == "Book C")
                    items.InnerList.Add(new Book { ID = 3, BookName = "Book C", ISBN = new Guid("33333333-3333-3333-3333-333333333333"), Sales = 5, PublishDate = new DateTime(2013, 4, 1, 0, 0, 0, 0) });
                else if (bookName == "Book D")
                    items.InnerList.Add(new Book { ID = 4, BookName = "Book D", ISBN = new Guid("44444444-4444-4444-4444-444444444444"), Sales = 1, PublishDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            return items;
        }

        
        ////////////////////////////////////////
        //Examples of custom reduce function
        ////////////////////////////////////////
        [CustomReduceFunction]
        private static HashSet<MapReduce.Row> PublishDateFilter(HashSet<MapReduce.Row> currentData, HashSet<MapReduce.Row> newData)
        {
            currentData.UnionWith(newData);

            //dynamic required in order to compile (will be resolved at runtime)
            Predicate<dynamic> p = a => a.PublishDate < new DateTime(2013, 4, 1, 0, 0, 0, 0);
            currentData.RemoveWhere(p);

            return currentData;
        }
    }

    #endregion

    
    #region Alarms object - Used for testing

    [Serializable()]
    public class Alarm
    {
        public int AlarmID { get; set; }
        public string Description { get; set; }
        public Guid AlarmGuid { get; set; }
        public double AlarmDuration { get; set; }
        public string Severity { get; set; }
        public int Count { get; set; }
        public DateTime CreateDate { get; set; }
    }

    [Serializable()]
    public class Alarms : ReadOnlyCollectionBase
    {
        ////////////////////////////////////////
        //Example to test operations (like union, sum, etc) and sorting
        ////////////////////////////////////////
        [PathInfo(PathInfo = "/Alarms", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms Get()
        {
            Alarms items = new Alarms();
            if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReducePortal"))
            {
                ////throw new Exception("test error on LocalNode...");
                items.InnerList.Add(new Alarm { AlarmID = 100, Description = "alarm1", AlarmGuid = new Guid("11111111-1111-1111-1111-111111111111"), Severity = "Critical", AlarmDuration = 1.1, Count = 1, CreateDate = new DateTime(2011, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Alarm { AlarmID = 2, Description = "alarm2", AlarmGuid = new Guid("22222222-2222-2222-2222-222222222222"), Severity = "Warning", AlarmDuration = 2.2, Count = 1, CreateDate = new DateTime(2012, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Alarm { AlarmID = 3, Description = "alarm3", AlarmGuid = new Guid("33333333-3333-3333-3333-333333333333"), Severity = "Major", AlarmDuration = 4.4, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData1"))
            {
                ////throw new Exception("test error on Node1...");
                items.InnerList.Add(new Alarm { AlarmID = 4, Description = "alarm4", AlarmGuid = new Guid("11111111-1111-1111-1111-111111111111"), Severity = "Critical", AlarmDuration = 4.4, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData2"))
            {
                ////throw new Exception("test error on Node2...");
                items.InnerList.Add(new Alarm { AlarmID = 4, Description = "alarm5", AlarmGuid = new Guid("22222222-2222-2222-2222-222222222222"), Severity = "Critical", AlarmDuration = 5.5, Count = 1, CreateDate = new DateTime(2016, 4, 1, 0, 0, 0, 0) });
            }
            else if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData3"))
            {
                ////throw new Exception("test error on Node3...");
                items.InnerList.Add(new Alarm { AlarmID = 4, Description = "alarm4", AlarmGuid = new Guid("33333333-3333-3333-3333-333333333333"), Severity = "Critical", AlarmDuration = 4.4, Count = 1, CreateDate = new DateTime(2017, 4, 1, 0, 0, 0, 0) });
                items.InnerList.Add(new Alarm { AlarmID = 4, Description = "alarm6", AlarmGuid = new Guid("11111111-1111-1111-1111-111111111111"), Severity = "Warning", AlarmDuration = 6.6, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            }
            return items;
        }


        ////////////////////////////////////////
        //Examples of custom reduce function
        ////////////////////////////////////////
        [CustomReduceFunction]
        private static HashSet<MapReduce.Row> AlarmDurationThreshold(HashSet<MapReduce.Row> currentData, HashSet<MapReduce.Row> newData)
        {
            currentData.UnionWith(newData);

            //dynamic required in order to compile (will be resolved at runtime)
            Predicate<dynamic> p = a => a.AlarmDuration > 4.0;
            currentData.RemoveWhere(p);

            return currentData;
        }

        [CustomReduceFunction]
        private static HashSet<MapReduce.Row> AlarmSeverityFilter(HashSet<MapReduce.Row> currentData, HashSet<MapReduce.Row> newData)
        {
            currentData.UnionWith(newData);

            //dynamic required in order to compile (will be resolved at runtime)
            Predicate<dynamic> p = a => a.Severity == "Warning";
            currentData.RemoveWhere(p);

            return currentData;
        }

        
        ////////////////////////////////////////
        //Examples to test URITemplate matches
        ////////////////////////////////////////
        [PathInfo(PathInfo = "/Alarms/{alarmID}", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms Get(int alarmID)
        {
            Alarms items = new Alarms();
            items.InnerList.Add(new Alarm { AlarmID = alarmID, Description = "alarm1", AlarmGuid = new Guid("11111111-1111-1111-1111-111111111111"), Severity = "Critical", AlarmDuration = 1.1, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            return items;
        }

        [PathInfo(PathInfo = "/Alarms/{alarmID}/{description}", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms Get(int alarmID, string description)
        {
            Alarms items = new Alarms();
            items.InnerList.Add(new Alarm { AlarmID = alarmID, Description = description, AlarmGuid = new Guid("22222222-2222-2222-2222-222222222222"), Severity = "Warning", AlarmDuration = 3.3, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            return items;
        }

        [PathInfo(PathInfo = "/Alarms/{alarmID}/{description}/{alarmGuid}", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms Get(int alarmID, string description, Guid alarmGuid)
        {
            Alarms items = new Alarms();
            items.InnerList.Add(new Alarm { AlarmID = alarmID, Description = description, AlarmGuid = alarmGuid, Severity = "Critical", AlarmDuration = 4.4, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            return items;
        }

        [PathInfo(PathInfo = "/Alarms/{alarmID}/{description}/{alarmGuid}/{duration}", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms Get(int alarmID, string description, Guid alarmGuid, double duration)
        {
            Alarms items = new Alarms();
            items.InnerList.Add(new Alarm { AlarmID = alarmID, Description = description, AlarmGuid = alarmGuid, Severity = "Major", AlarmDuration = duration, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            
            if (Assembly.GetExecutingAssembly().CodeBase.Contains("MapReduceData3"))
                throw new Exception("test error on Node3...");

            return items;
        }

        [PathInfo(PathInfo = "/Alarms/GetByGuid/{alarmGuid}", ReturnItemType = "BusinessObjects.Alarm")]
        public static Alarms GetByGuid(Guid alarmGuid)
        {
            Alarms items = new Alarms();
            items.InnerList.Add(new Alarm { AlarmID = 2, Description = "alarm2", AlarmGuid = alarmGuid, Severity = "Critical", AlarmDuration = 2.2, Count = 1, CreateDate = new DateTime(2014, 4, 1, 0, 0, 0, 0) });
            return items;
        }
    }

    #endregion

    
    #region CheeseBurgerWorkflow object - Used for testing workflow

    [Serializable()]
    public class KeyValuePair
    {
        public object Key { get; set; }
        public object Value { get; set; }
    }
    [Serializable()]
    public class CheeseBurgerWorkflow
    {
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step1", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step1()
        {
            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step1", Value = "Bottom Bun" });
            return step;
        }
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step2", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step2()
        {
            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step2", Value = "Hamburger Patty" });
            return step;
        }
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step3", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step3()
        {
            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step3", Value = "Cheese" });
            return step;
        }
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step4", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step4()
        {
            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step4", Value = "Lettuce" });
            return step;
        }
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step5", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step5()
        {
            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step5", Value = "Tomato" });
            return step;
        }
        [PathInfo(PathInfo = "/CheeseBurgerWorkflow/Step6", ReturnItemType = "BusinessObjects.KeyValuePair")]
        public static HashSet<KeyValuePair> Step6()
        {
            ////throw new Exception("test error on Step6...");

            //do some work for this workflow step
            HashSet<KeyValuePair> step = new HashSet<KeyValuePair>();
            step.Add(new KeyValuePair { Key = "Step6", Value = "Top Bun" });
            return step;
        }
    }

    #endregion

    
    #region Famous "Hello World" Example - Used for testing

    [Serializable()]
    public class Utilities
    {
        [PathInfo(PathInfo = "/WordCount/{documentName}", ReturnItemType = "MapReduce.Row")]
        public static HashSet<MapReduce.Row> GetWordCount(string documentName)
        {
            HashSet<MapReduce.Row> rows = new HashSet<MapReduce.Row>();
            //Note: Need to put Test.txt file in each node's virtual directory next to the web.config file (NOT in \bin directory).
            byte[] bytes = File.ReadAllBytes(AppDomain.CurrentDomain.BaseDirectory + documentName);
            string text = Encoding.ASCII.GetString(bytes);
            string[] words = text.Split(new char[] { ' ', '\r', '\n' });
            foreach (string word in words)
            {
                dynamic row = new MapReduce.Row();
                row["Word"] = word;
                row["Count"] = 1;
                rows.Add(row);
            }
            return rows;
        }
    }
    #endregion

}

#endregion
